@@ -12,6 +12,7 @@
 #include <fcntl.h>
 #include <dirent.h>
 #include <inttypes.h>
+#include <time.h>
 #include <sys/wait.h>
 #include <sys/stat.h>
 #include "dr.h"
@@ -21,7 +22,7 @@
 static int block_sz_bytes = 0;
 static char *state_strs[] = {"offline", "online"};
 
-static char *usagestr = "-c mem {-a | -r} {-q <quantity> -p {variable_weight | ent_capacity} | {-q <quantity> | s <drc_name>}}";
+static char *usagestr = "-c mem {-a | -r} {-q <quantity> -p {variable_weight | ent_capacity} | {-q <quantity> | -s [<drc_name> | <drc_index>]}}";
 
 /**
  * mem_usage
@@ -105,10 +106,8 @@ get_mem_scns(struct dr_node *lmb)
 		struct stat sbuf;
 
 		scn = zalloc(sizeof(*scn));
-		if (scn == NULL) {
-			err_msg("Could not allocate memory section\n.");
+		if (scn == NULL)
 			return -1;
-		}
 
 		sprintf(scn->sysfs_path, sysfs_path, mem_scn);
 		scn->phys_addr = phys_addr;
@@ -154,7 +153,8 @@ get_lmb_size(struct dr_node *lmb)
 
 	rc = get_property(lmb->ofdt_path, "reg", &regs, sizeof(regs));
 	if (rc) {
-		dbg("Could not determine lmb size for %s\n", lmb->ofdt_path);
+		say(DEBUG, "Could not determine LMB size for %s\n",
+		    lmb->ofdt_path);
 		return rc;
 	}
 
@@ -178,8 +178,11 @@ get_mem_node_lmbs(struct lmb_list_head *lmb_list)
 	int rc = 0;
 
 	d = opendir(OFDT_BASE);
-	if (d == NULL)
+	if (d == NULL) {
+		report_unknown_error(__FILE__, __LINE__);
+		say(DEBUG, "Could not open %s\n", OFDT_BASE);
 		return -1;
+	}
 
 	while ((de = readdir(d)) != NULL) {
 		char path[1024];
@@ -204,7 +207,7 @@ get_mem_node_lmbs(struct lmb_list_head *lmb_list)
 		}
 
 		if (lmb == NULL) {
-			dbg("Could not find lmb with drc-index of %x\n",
+			say(DEBUG, "Could not find LMB with drc-index of %x\n",
 			    lmb->drc_index);
 			rc = -1;
 			break;
@@ -214,14 +217,18 @@ get_mem_node_lmbs(struct lmb_list_head *lmb_list)
 		lmb->is_owned = 1;
 
 		/* Find the lmb size for this lmb */
+		/* XXX Do nothing with size and break if it can't be found
+		 * but don't change rc to indicate failure?
+		 * Why not continue? If we break, set rc=-1? */
 		if (get_lmb_size(lmb))
 			break;
 
 		/* Find the physical address for this lmb */
 		tmp = strrchr(lmb->ofdt_path, '@');
 		if (tmp == NULL) {
-			dbg("Could not determine physical address for %s\n",
-			    lmb->ofdt_path);
+			say(DEBUG, "Could not determine physical address for "
+			    "%s\n", lmb->ofdt_path);
+			/* XXX No rc change? */
 			break;
 		}
 
@@ -257,7 +264,7 @@ get_dynamic_reconfig_lmbs(struct lmb_list_head *lmb_list)
 	rc = get_property(DYNAMIC_RECONFIG_MEM, "ibm,lmb-size",
 			  &lmb_sz, sizeof(lmb_sz));
 	if (rc) {
-		dbg("Could not retrieve drconf lmb size\n");
+		say(DEBUG, "Could not retrieve drconf LMB size\n");
 		return rc;
 	}
 
@@ -265,15 +272,15 @@ get_dynamic_reconfig_lmbs(struct lmb_list_head *lmb_list)
 						   "ibm,dynamic-memory");
 	lmb_list->drconf_buf = zalloc(lmb_list->drconf_buf_sz);
 	if (lmb_list->drconf_buf == NULL) {
-		dbg("Could not allocate buffer to get dynamic reconfigurable "
-		    "memory\n");
+		say(DEBUG, "Could not allocate buffer to get dynamic "
+		    "reconfigurable memory\n");
 		return -1;
 	}
 
 	rc = get_property(DYNAMIC_RECONFIG_MEM, "ibm,dynamic-memory",
 			  lmb_list->drconf_buf, lmb_list->drconf_buf_sz);
 	if (rc) {
-		dbg("Could not retrieve dynamic reconfigurable memory "
+		say(DEBUG, "Could not retrieve dynamic reconfigurable memory "
 		    "property\n");
 		return -1;
 	}
@@ -293,7 +300,7 @@ get_dynamic_reconfig_lmbs(struct lmb_list_head *lmb_list)
 		}
 
 		if (lmb == NULL) {
-			dbg("Could not find lmb with drc-index of %x\n",
+			say(DEBUG, "Could not find LMB with drc-index of %x\n",
 			    lmb->drc_index);
 			rc = -1;
 			break;
@@ -316,10 +323,51 @@ get_dynamic_reconfig_lmbs(struct lmb_list_head *lmb_list)
 		drmem++; /* trust your compiler */
 	}
 
-	dbg("Found %d owning lmbs\n", found);
+	say(INFO, "Found %d LMBs currently allocated\n", found);
 	return rc;
 }
 
+/**
+ * shuffle_lmbs
+ * @brief Randomly shuffle list of lmbs
+ *
+ * @param list pointer to lmb list to be randomly shuffled
+ * @param length number of lmbs in the list
+ *
+ * @return list of shuffled lmbs
+ */
+struct dr_node *
+shuffle_lmbs(struct dr_node *lmb_list, int length)
+{
+	struct dr_node **shuffled_lmbs, *lmb;
+	int i, j;
+	
+	srand(time(NULL));
+
+	shuffled_lmbs = zalloc(sizeof(*shuffled_lmbs) * length);
+
+	for (i = 0, lmb = lmb_list; lmb; i++, lmb = lmb->next) {
+		j = rand() % (i + 1);
+
+		if (j == i) {
+			shuffled_lmbs[i] = lmb;
+		} else {
+			shuffled_lmbs[i] = shuffled_lmbs[j];
+			shuffled_lmbs[j] = lmb;
+		}
+	}
+
+	for (i = 0; i < (length - 1); i++)
+		shuffled_lmbs[i]->next = shuffled_lmbs[i + 1];
+
+	shuffled_lmbs[length - 1]->next = NULL;
+
+	lmb = shuffled_lmbs[0];
+	free(shuffled_lmbs);
+
+	return lmb;
+}
+
 /**
  * get_lmbs
  * @brief Build a list of all possible lmbs for the system
@@ -340,12 +388,14 @@ get_lmbs(unsigned int sort)
 	int found = 0;
 
 	drc_list = get_drc_info(OFDT_BASE);
-	if (drc_list == NULL)
+	if (drc_list == NULL) {
+		report_unknown_error(__FILE__, __LINE__);
 		return NULL;
+	}
 
 	lmb_list = zalloc(sizeof(*lmb_list));
 	if (lmb_list == NULL) {
-		dbg("Could not allocate lmb list head\n");
+		say(DEBUG, "Could not allocate LMB list head\n");
 		return NULL;
 	}
 
@@ -377,12 +427,16 @@ get_lmbs(unsigned int sort)
 		found++;
 	}
 
-	dbg("Found %d lmbs\n", found);
+	if (sort == LMB_RANDOM_SORT)
+		lmb_list->lmbs = shuffle_lmbs(lmb_list->lmbs, found);
+
+	say(INFO, "Maximum of %d LMBs\n", found);
 
 	rc = get_str_attribute("/sys/devices/system/memory",
 			       "/block_size_bytes", &buf, DR_STR_MAX);
 	if (rc) {
-		dbg("Could not determine block size bytes for memory.\n");
+		say(DEBUG,
+		    "Could not determine block size bytes for memory.\n");
 		free_lmbs(lmb_list);
 		return NULL;
 	}
@@ -430,6 +484,7 @@ get_lmbs(unsigned int sort)
 static struct dr_node *
 get_available_lmb(struct options *opts, struct dr_node *start_lmb)
 {
+	uint32_t drc_index;
 	struct dr_node *lmb;
 	struct dr_node *usable_lmb = NULL;
 	int balloon_active = ams_balloon_active();
@@ -437,9 +492,13 @@ get_available_lmb(struct options *opts, struct dr_node *start_lmb)
 	for (lmb = start_lmb; lmb; lmb = lmb->next) {
 		int rc;
 
-		if (opts->usr_drc_name)
-			if (strcmp(lmb->drc_name, opts->usr_drc_name))
+		if (opts->usr_drc_name) {
+			drc_index = strtoul(opts->usr_drc_name, NULL, 0);
+
+			if ((strcmp(lmb->drc_name, opts->usr_drc_name))
+			    && (lmb->drc_index != drc_index))
 				continue;
+		}
 
 		if (lmb->unusable)
 			continue;
@@ -471,10 +530,8 @@ get_available_lmb(struct options *opts, struct dr_node *start_lmb)
 	}
 
 	if (usable_lmb)
-		dbg("Found available lmb, %s, drc index 0x%x\n",
+		say(DEBUG, "Found available LMB, %s, drc index 0x%x\n",
 		    usable_lmb->drc_name, usable_lmb->drc_index);
-	else
-		dbg("Could not find available lmb\n");
 
 	return usable_lmb;
 }
@@ -537,7 +594,8 @@ update_drconf_node(struct dr_node *lmb, struct lmb_list_head *lmb_list,
 	rc = get_phandle(lmb->ofdt_path, &phandle);
 
 	if (rc) {
-		dbg("Failed to get phandle: %d\n",rc);
+		say(DEBUG, "Failed to get phandle for %s under %s. (rc=%d)\n", 
+				lmb->drc_name, lmb->ofdt_path, rc);
 		return rc;
 	}
 
@@ -600,7 +658,8 @@ add_device_tree_lmb(struct dr_node *lmb, struct lmb_list_head *lmb_list)
 			 * In this case we provide this little hack to allow
 			 * memory add to work on these kernels.
 			 */
-			dbg("Assuming older kernel, trying to add node\n");
+			say(DEBUG, "Assuming older kernel, trying to add "
+			    "node\n");
 			lmb->lmb_of_node = configure_connector(lmb->drc_index);
 			if (lmb->lmb_of_node == NULL) {
 				release_drc(lmb->drc_index, MEM_DEV);
@@ -636,8 +695,8 @@ add_device_tree_lmb(struct dr_node *lmb, struct lmb_list_head *lmb_list)
 		 */
 		char *tmp = strrchr(lmb->ofdt_path, '@');
 		if (tmp == NULL) {
-			dbg("Could not determine physical address for %s\n",
-			    lmb->ofdt_path);
+			say(DEBUG, "Could not determine physical address for "
+			    "%s\n", lmb->ofdt_path);
 			remove_device_tree_nodes(lmb->ofdt_path);
 			return -1;
 		}
@@ -678,8 +737,8 @@ get_mem_scn_state(struct mem_scn *mem_scn)
 
 	file = open(path, O_RDONLY);
 	if (file <= 0) {
-		dbg("Could not open %s to validate its state.\n%s\n",
-			    path, strerror(errno));
+		say(DEBUG, "Could not open %s to validate its state.\n%s\n",
+		    path, strerror(errno));
 		return -1;
 	}
 
@@ -711,14 +770,19 @@ set_mem_scn_state(struct mem_scn *mem_scn, int state)
 	char path[DR_PATH_MAX];
 	int rc = 0;
 	int unused;
+	time_t t;
+	char tbuf[128];
 
+	time(&t);
+	strftime(tbuf, 128, "%T", localtime(&t));
 	memset(path, 0, DR_PATH_MAX);
 	sprintf(path, "%s/state", mem_scn->sysfs_path);
-	dbg("Marking %s %s\n", mem_scn->sysfs_path, state_strs[state]);
+	say(DEBUG, "%s Marking %s %s\n", tbuf, mem_scn->sysfs_path,
+			state_strs[state]);
 
 	file = open(path, O_WRONLY);
 	if (file <= 0) {
-		dbg("Could not open %s to %s memory.\n%s\n",
+		say(DEBUG, "Could not open %s to %s memory.\n\t%s\n",
 		    path, state_strs[state], strerror(errno));
 		close(file);
 		return -1;
@@ -728,9 +792,16 @@ set_mem_scn_state(struct mem_scn *mem_scn, int state)
 	close(file);
 
 	if (get_mem_scn_state(mem_scn) != state) {
-		dbg("Could not %s %s.\n", state_strs[state],
+		time(&t);
+		strftime(tbuf, 128, "%T", localtime(&t));
+		say(DEBUG, "%s Could not %s %s.\n", tbuf, state_strs[state],
 		    mem_scn->sysfs_path);
 		rc = EAGAIN;
+	} else {
+		time(&t);
+		strftime(tbuf, 128, "%T", localtime(&t));
+		say(DEBUG, "%s Completed marking %s %s.\n", tbuf,
+				mem_scn->sysfs_path, state_strs[state]);
 	}
 
 	return rc;
@@ -752,7 +823,8 @@ probe_lmb(struct dr_node *lmb)
 
 	probe_file = open(MEM_PROBE_FILE, O_WRONLY);
 	if (probe_file <= 0) {
-		dbg("Could not open %s to probe memory\n", MEM_PROBE_FILE);
+		say(DEBUG, "Could not open %s to probe memory\n",
+		    MEM_PROBE_FILE);
 		return errno;
 	}
 
@@ -762,10 +834,10 @@ probe_lmb(struct dr_node *lmb)
 		memset(addr, 0, DR_STR_MAX);
 		sprintf(addr, "0x%"PRIx64, scn->phys_addr);
 
-		dbg("Probing memory address 0x%llx\n", scn->phys_addr);
+		say(DEBUG, "Probing memory address 0x%llx\n", scn->phys_addr);
 		rc = write(probe_file, addr, strlen(addr));
 		if (rc == -1) {
-			dbg("Probe failed:\n%s\n", strerror(errno));
+			say(DEBUG, "Probe failed:\n%s\n", strerror(errno));
 			return rc;
 		}
 	}
@@ -788,7 +860,7 @@ set_lmb_state(struct dr_node *lmb, int state)
 	int rc = 0;
 	struct stat sbuf;
 
-	dbg("Attempting to %s lmb.\n", state_strs[state]);
+	say(INFO, "Attempting to %s %s.\n", state_strs[state], lmb->drc_name);
 
 	if (state == ONLINE) {
 		rc = probe_lmb(lmb);
@@ -820,8 +892,15 @@ set_lmb_state(struct dr_node *lmb, int state)
 		}
 	}
 
-	if (rc == EAGAIN)
-		err_msg("Could not %s lmb.\n", state_strs[state]);
+	if (rc) {
+		if (rc == EAGAIN)
+			say(INFO, "Could not %s %s at this time.\n",
+				      state_strs[state], lmb->drc_name);
+		else
+			report_unknown_error(__FILE__, __LINE__);
+
+	} else
+		say(INFO, "%s is %s.\n", lmb->drc_name, state_strs[state]);
 
 	return rc;
 }
@@ -847,6 +926,9 @@ add_lmbs(struct options *opts, struct lmb_list_head *lmb_list)
 
 	lmb_list->lmbs_modified = 0;
 	while (lmb_list->lmbs_modified < (int)opts->quantity) {
+		if (drmgr_timed_out())
+			break;
+
 		lmb = get_available_lmb(opts, lmb_head);
 		if (lmb == NULL)
 			return -1;
@@ -856,12 +938,14 @@ add_lmbs(struct options *opts, struct lmb_list_head *lmb_list)
 
 		rc = acquire_drc(lmb->drc_index);
 		if (rc) {
+			report_unknown_error(__FILE__, __LINE__);
 			lmb->unusable = 1;
 			continue;
 		}
 
 		rc = add_device_tree_lmb(lmb, lmb_list);
 		if (rc) {
+			report_unknown_error(__FILE__, __LINE__);
 			release_drc(lmb->drc_index, MEM_DEV);
 			lmb->unusable = 1;
 			continue;
@@ -869,6 +953,7 @@ add_lmbs(struct options *opts, struct lmb_list_head *lmb_list)
 
 		rc = set_lmb_state(lmb, ONLINE);
 		if (rc) {
+			report_unknown_error(__FILE__, __LINE__);
 			remove_device_tree_lmb(lmb, lmb_list);
 			release_drc(lmb->drc_index, MEM_DEV);
 			lmb->unusable = 1;
@@ -895,13 +980,16 @@ mem_add(struct options *opts)
 	int rc;
 
 	lmb_list = get_lmbs(LMB_NORMAL_SORT);
-	if (lmb_list == NULL)
+	if (lmb_list == NULL) {
+		say(ERROR, "Could not gather LMB (logical memory block "
+				"information.\n");
 		return -1;
+	}
 
-	dbg("Adding %d lmbs\n", opts->quantity);
+	say(DEBUG, "Attempting to add %d LMBs\n", opts->quantity);
 	rc = add_lmbs(opts, lmb_list);
 
-	dbg("Added %d of %d requested lmb(s)\n", lmb_list->lmbs_modified,
+	say(DEBUG, "Added %d of %d requested LMB(s)\n", lmb_list->lmbs_modified,
 	    opts->quantity);
 	printf("DR_TOTAL_RESOURCES=%d\n", lmb_list->lmbs_modified);
 
@@ -925,6 +1013,9 @@ remove_lmbs(struct options *opts, struct lmb_list_head *lmb_list)
 	int rc;
 
 	while (lmb_list->lmbs_modified < (int)opts->quantity) {
+		if (drmgr_timed_out())
+			break;
+
 		lmb = get_available_lmb(opts, lmb_head);
 		if (!lmb)
 			return -1;
@@ -940,6 +1031,7 @@ remove_lmbs(struct options *opts, struct lmb_list_head *lmb_list)
 
 		rc = remove_device_tree_lmb(lmb, lmb_list);
 		if (rc) {
+			report_unknown_error(__FILE__, __LINE__);
 			set_lmb_state(lmb, ONLINE);
 			lmb->unusable = 1;
 			continue;
@@ -953,6 +1045,7 @@ remove_lmbs(struct options *opts, struct lmb_list_head *lmb_list)
 
 		rc = release_drc(lmb->drc_index, 0);
 		if (rc) {
+			report_unknown_error(__FILE__, __LINE__);
 			add_device_tree_lmb(lmb, lmb_list);
 			set_lmb_state(lmb, ONLINE);
 			lmb->unusable = 1;
@@ -978,11 +1071,15 @@ mem_remove(struct options *opts)
 	struct lmb_list_head *lmb_list;
 	struct dr_node *lmb;
 	unsigned int removable = 0;
+	unsigned int requested = opts->quantity;
 	int rc = 0;
 
-	lmb_list = get_lmbs(LMB_REVERSE_SORT);
-	if (lmb_list == NULL)
+	lmb_list = get_lmbs(LMB_RANDOM_SORT);
+	if (lmb_list == NULL) {
+		say(ERROR, "Could not gather LMB (logical memory block "
+				"information.\n");
 		return -1;
+	}
 
 	/* Can not know which lmbs are removable by the is_removable field
 	 * if AMS ballooning is active.
@@ -996,25 +1093,29 @@ mem_remove(struct options *opts)
 				removable++;
 		}
 
-		if (removable < opts->quantity) {
-			err_msg("There is not enough removable memory available to "
-				"fulfill the request.\n");
+		if (removable == 0) {
+			say(ERROR, "There is not enough removable memory "
+			    "available to fulfill the request.\n");
 			rc = -1;
 		}
+
+		if (removable < opts->quantity) {
+			say(INFO, "Only %u LMBs are currently candidates "
+					"for removal.\n", removable);
+			opts->quantity = removable;
+		}
 	}
 
 	if (!rc) {
-		dbg("Removing %d lmbs\n", opts->quantity);
+		say(DEBUG, "Attempting removal of %d LMBs\n", opts->quantity);
 		rc = remove_lmbs(opts, lmb_list);
-		if (rc)
-			dbg("Memory removal request failed\n");
 	}
 
-	dbg("Removed %d of %d requested lmb(s)\n", lmb_list->lmbs_modified,
-	    opts->quantity);
-	if (lmb_list->lmbs_modified < opts->quantity)
-		dbg("Unable to hotplug remove the remaining %d lmb(s)\n",
-		    opts->quantity - lmb_list->lmbs_modified);
+	say(ERROR, "Removed %d of %d requested LMB(s)\n",
+	    lmb_list->lmbs_modified, requested);
+	if (lmb_list->lmbs_modified < requested)
+		say(ERROR, "Unable to hotplug remove the remaining %d LMB(s)\n",
+		    requested - lmb_list->lmbs_modified);
 	printf("DR_TOTAL_RESOURCES=%d\n", lmb_list->lmbs_modified);
 
 	free_lmbs(lmb_list);
@@ -1065,8 +1166,8 @@ ehea_compatable(int action)
 		/* File doesn't exist, memory dlpar operations are not
 		 * supported by this version of the ehea driver.
 		 */
-		err_msg("The eHEA module for this system does not support "
-			"memory DLPAR operations.\n");
+		say(INFO, "The eHEA module for this system does not support "
+		    "memory DLPAR operations.\n");
 		return 0;
 	}
 
@@ -1083,9 +1184,9 @@ ehea_compatable(int action)
 		rc = 1;
 
 	if (!rc)
-		err_msg("The eHEA modules loaded on this system does not "
-			"support memory DLPAR %s operations.\n",
-			(action == ADD) ? "add" : "remove");
+		say(INFO, "The eHEA modules loaded on this system does not "
+		    "support memory DLPAR %s operations.\n",
+		    (action == ADD) ? "add" : "remove");
 	return rc;
 }
 
@@ -1109,15 +1210,12 @@ drslot_chrp_mem(struct options *opts)
 		return update_sysparm(opts);
 	}
 
-	if (! mem_dlpar_capable()) {
-		err_msg("DLPAR memory operations are not supported on"
-			"this kernel.");
+	if (! mem_dlpar_capable() || ! ehea_compatable(opts->action)) {
+		say(ERROR, "DLPAR memory operations are not supported on"
+		    "this kernel.");
 		return -1;
 	}
 
-	if (! ehea_compatable(opts->action))
-		return -1;
-
 	/* The recursive nature of the routines that add/remove lmbs
 	 * require that the quantity be non-zero.
 	 */
