@@ -32,7 +32,7 @@ static char *speed_error = "Add operation failed. The 33MHz PCI card may\n"
 #define USER_QUIT  0		/* user wants to bail out of operation	 */
 #define USER_CONT  1		/* user wants to continue with operation */
 
-static char *usagestr = 	"-c pci -s slot {-i | -a [-I] | -r [-I] | -R [-I]}";
+static char *usagestr = 	"-c pci -s <drc_name | drc_index> {-i | -a [-I] | -r [-I] | -R [-I]}";
 /**
  * pci_usage
  *
@@ -61,9 +61,9 @@ process_led(struct dr_node *node, int led)
 	rtas_rc = rtas_set_indicator(DR_INDICATOR, node->drc_index, led);
 	if (rtas_rc) {
 		if (rtas_rc == HW_ERROR)
-			err_msg("%s", hw_error);
+			say(ERROR, "%s", hw_error);
 		else
-			err_msg("%s", sw_error);
+			say(ERROR, "%s", sw_error);
 
 		return -1;
 	}
@@ -96,6 +96,28 @@ identify_slot(struct dr_node *node)
 		return (USER_QUIT);
 }
 
+static char *
+find_drc_name(uint32_t drc_index, struct dr_node *all_nodes)
+{
+	struct dr_node *node;
+
+	node = all_nodes;
+	while (node != NULL) {
+		say(DEBUG, "%#x =? %#x\n", drc_index, node->drc_index);
+		if (node->drc_index == drc_index) {
+			say(DEBUG, "Found drc_name %s\n", node->drc_name);
+			return node->drc_name;
+		} else
+			node = node->next;
+	}
+
+	if ((node == NULL))
+		say(ERROR, "Could not find drc_name for index %#x\n", drc_index);
+
+	return NULL;
+}
+
+
 /**
  * find_slot
  *
@@ -127,12 +149,43 @@ find_slot(char *drc_name, struct dr_node *all_nodes)
 	}
 
 	if ((node == NULL) || (node->skip))
-		err_msg("The specified PCI slot is either invalid\n"
-			"or does not support hot plug operations.\n");
+		say(ERROR, "The specified PCI slot is either invalid\n"
+		    "or does not support hot plug operations.\n");
 
 	return node;
 }
 
+static int check_card_presence(struct options *opts, struct dr_node *node)
+{
+	int state = EMPTY;
+	int i, keep_working;
+
+	say(DEBUG, "Waiting for the card to appear...\n");
+	do {
+		keep_working = 0;
+
+		for (i = 0; i < 30; i++) {
+			state = dr_entity_sense(node->drc_index);
+			if (state != EMPTY)
+				return state;
+
+			sleep(1);
+		}
+
+		if (0 == opts->noprompt) {
+			printf("The card still does not appear to be present"
+			       "\npress Enter to continue to wait or enter "
+			       "'x' to exit.\n");
+
+			if ((getchar() == '\n'))
+				keep_working = 1;
+                }
+
+        } while (keep_working);
+
+	return state;
+}
+
 /**
  * card_present
  * @brief Determines if there is a card present in the specified slot
@@ -144,26 +197,27 @@ find_slot(char *drc_name, struct dr_node *all_nodes)
  * maybe isolate and try sensing again. If we hit fatal errors, call
  * error_exit to clean up and exit the command.
  *
+ * @param opts
  * @param slot pointer to slot we're checking
  * @param power_state state of power when we leave this routine
  * @param isolate_state state of isolation when we leave this routine
  * @returns EMPTY or PRESENT
  */
 static int
-card_present(struct dr_node *node, int *power_state, int *isolate_state)
+card_present(struct options *opts, struct dr_node *node, int *power_state,
+	     int *isolate_state)
 {
 	int state, rc;
 
 	*power_state = POWER_OFF;	/* initialize */
 	*isolate_state = ISOLATE;
 
-	state = dr_entity_sense(node->drc_index);
+	state = check_card_presence(opts, node);
 	if ((state == EMPTY) || (state == PRESENT))
-		/* if we get an answer, we're done */
 		return state;
 
 	else if (state == HW_ERROR) {
-		err_msg("%s", hw_error);
+		say(ERROR, "%s", hw_error);
 		return -1;
 	}
 	else if ((state == NEED_POWER) || (state == PWR_ONLY)) {
@@ -171,17 +225,17 @@ card_present(struct dr_node *node, int *power_state, int *isolate_state)
 		rc = set_power(node->drc_power, POWER_ON);
 		if (rc) {
 			if (rc == HW_ERROR) {
-				err_msg("%s", hw_error);
+				say(ERROR, "%s", hw_error);
 				return -1;
 			}
 			else if (rc == SPEED_ERROR) {
-				err_msg("%s", speed_error);
+				say(ERROR, "%s", speed_error);
 				set_power(node->drc_power, POWER_OFF);
 				return -1;
 			}
 			/* Catch any other errors.  */
 			else {
-				err_msg("%s", sw_error);
+				say(ERROR, "%s", sw_error);
 				set_power(node->drc_power, POWER_OFF);
 				return -1;
 			}
@@ -196,9 +250,9 @@ card_present(struct dr_node *node, int *power_state, int *isolate_state)
 						node->drc_index, UNISOLATE);
 			if (rc) {
 				if (rc == HW_ERROR)
-					err_msg("%s", hw_error);
+					say(ERROR, "%s", hw_error);
 				else
-					err_msg("%s", sw_error);
+					say(ERROR, "%s", sw_error);
 
 				rtas_set_indicator(ISOLATION_STATE,
 						   node->drc_index, ISOLATE);
@@ -211,15 +265,15 @@ card_present(struct dr_node *node, int *power_state, int *isolate_state)
 		/* Now we have power on, and the unisolate is done
 		 * if it was needed. check for card again.
 		 */
-		state = dr_entity_sense(node->drc_index);
+		state = check_card_presence(opts, node);
 		if ((state == EMPTY) || (state == PRESENT))
 			return state;
 
 		if (state) {
 			if (rc == HW_ERROR)
-				err_msg("%s", hw_error);
+				say(ERROR, "%s", hw_error);
 			else
-				err_msg("%s", sw_error);
+				say(ERROR, "%s", sw_error);
 
 			rtas_set_indicator(ISOLATION_STATE, node->drc_index,
 					   ISOLATE);
@@ -229,7 +283,7 @@ card_present(struct dr_node *node, int *power_state, int *isolate_state)
 	}
 	else {
   		/* catch any other errors from first dr_entity_sense */
-		err_msg("%s", sw_error);
+		say(ERROR, "%s", sw_error);
 	}
 
 	return state;
@@ -289,11 +343,12 @@ do_identify(struct options *opts, struct dr_node *all_nodes)
  * If the OF tree cannot be updated, the slot is powered
  * off, isolated, and the LED is turned off.
  *
+ * @param opts
  * @param slot
  * @returns 0 on success, !0 on failure
  */
 static int
-add_work(struct dr_node *node)
+add_work(struct options *opts, struct dr_node *node)
 {
 	int pow_state;	/* Tells us if power was turned on when	 */
 	int iso_state;	/* Tells us isolation state after 	 */
@@ -304,28 +359,28 @@ add_work(struct dr_node *node)
 	if (process_led(node, LED_ON))
 		return -1;
 
-	dbg("is calling card_present\n");
-	rc = card_present(node, &pow_state, &iso_state);
+	say(DEBUG, "is calling card_present\n");
+	rc = card_present(opts, node, &pow_state, &iso_state);
 	if (!rc) {
-		err_msg("No PCI card was detected in the specified "
-			"PCI slot.\n");
+		say(ERROR, "No PCI card was detected in the specified "
+		    "PCI slot.\n");
 		rtas_set_indicator(ISOLATION_STATE, node->drc_index, ISOLATE);
 		set_power(node->drc_power, POWER_OFF);
 		return -1;
 	}
 
 	if (!pow_state) {
-		dbg("is calling set_power(POWER_ON index 0x%x, "
+		say(DEBUG, "is calling set_power(POWER_ON index 0x%x, "
 		    "power_domain 0x%x\n", node->drc_index, node->drc_power);
 
 		rc = set_power(node->drc_power, POWER_ON);
 		if (rc) {
 			if (rc == HW_ERROR)
-				err_msg("%s", hw_error);
+				say(ERROR, "%s", hw_error);
 			else if (rc == SPEED_ERROR)
-				err_msg("%s", speed_error);
+				say(ERROR, "%s", speed_error);
 			else
-				err_msg("%s", sw_error);
+				say(ERROR, "%s", sw_error);
 
 			rtas_set_indicator(ISOLATION_STATE, node->drc_index,
 					   ISOLATE);
@@ -335,16 +390,16 @@ add_work(struct dr_node *node)
 	}
 
 	if (!iso_state) {
-		dbg("calling rtas_set_indicator(UNISOLATE index 0x%x)\n",
+		say(DEBUG, "calling rtas_set_indicator(UNISOLATE index 0x%x)\n",
 		    node->drc_index);
 
 		rc = rtas_set_indicator(ISOLATION_STATE, node->drc_index,
 					UNISOLATE);
 		if (rc) {
 			if (rc == HW_ERROR)
-				err_msg("%s", hw_error);
+				say(ERROR, "%s", hw_error);
 			else
-				err_msg("%s", sw_error);
+				say(ERROR, "%s", sw_error);
 
 			rtas_set_indicator(ISOLATION_STATE, node->drc_index,
 					   ISOLATE);
@@ -364,11 +419,12 @@ add_work(struct dr_node *node)
 		return -1;
 	}
 
-	dbg("Adding %s to %s\n", new_nodes->name, node->ofdt_path);
+	say(DEBUG, "Adding %s to %s\n", new_nodes->name, node->ofdt_path);
 	rc = add_device_tree_nodes(node->ofdt_path, new_nodes);
 	if (rc) {
-		dbg("add_device_tree_nodes failed at %s\n", node->ofdt_path);
-		err_msg("%s", sw_error);
+		say(DEBUG, "add_device_tree_nodes failed at %s\n",
+		    node->ofdt_path);
+		say(ERROR, "%s", sw_error);
 		rtas_set_indicator(ISOLATION_STATE, node->drc_index, ISOLATE);
 		set_power(node->drc_power, POWER_OFF);
 		return -1;
@@ -422,8 +478,7 @@ do_add(struct options *opts, struct dr_node *all_nodes)
 		 * LED on and exit with user error.
 		 */
 		process_led(node, LED_ON);
-		err_msg("The specified PCI slot is already "
-			"occupied.\n");
+		say(ERROR, "The specified PCI slot is already occupied.\n");
 		return -1;
 	}
 
@@ -433,29 +488,29 @@ do_add(struct options *opts, struct dr_node *all_nodes)
 	 * card in.
 	 */
 
-	dbg("is calling rtas_set_indicator(ISOLATE index 0x%x)\n",
+	say(DEBUG, "is calling rtas_set_indicator(ISOLATE index 0x%x)\n",
 	    node->drc_index);
 
 	rc = rtas_set_indicator(ISOLATION_STATE, node->drc_index, ISOLATE);
 	if (rc) {
 		if (rc == HW_ERROR)
-			err_msg("%s", hw_error);
+			say(ERROR, "%s", hw_error);
 		else
-			err_msg("%s", sw_error);
+			say(ERROR, "%s", sw_error);
 
 		set_power(node->drc_power, POWER_OFF);
 		return -1;
 	}
 
-	dbg("is calling set_power(POWER_OFF index 0x%x, "
+	say(DEBUG, "is calling set_power(POWER_OFF index 0x%x, "
 	    "power_domain 0x%x) \n", node->drc_index, node->drc_power);
 
 	rc = set_power(node->drc_power, POWER_OFF);
 	if (rc) {
 		if (rc == HW_ERROR)
-			err_msg("%s", hw_error);
+			say(ERROR, "%s", hw_error);
 		else
-			err_msg("%s", sw_error);
+			say(ERROR, "%s", sw_error);
 
 		return -1;
 	}
@@ -482,14 +537,20 @@ do_add(struct options *opts, struct dr_node *all_nodes)
 	/* Call the routine which determines
 	 * what the user wants and does it.
 	 */
-	rc = add_work(node);
+	rc = add_work(opts, node);
 	if (rc)
 		return rc;
 
-	dbg("is calling enable_slot to config adapter\n");
+	say(DEBUG, "is calling enable_slot to config adapter\n");
 
-	/* Try to config the adapter */
-	set_hp_adapter_status(PHP_CONFIG_ADAPTER, node->drc_name);
+	/* Try to config the adapter. The rpaphp module doesn't play well with
+	 * qemu pci slots so we let the generic kernel pci code probe the device
+	 * by rescanning the bus in the qemu virtio case.
+	 */
+	if (!opts->pci_virtio)
+		set_hp_adapter_status(PHP_CONFIG_ADAPTER, node->drc_name);
+	else
+		pci_rescan_bus();
 
 	return 0;
 }
@@ -521,7 +582,7 @@ remove_work(struct options *opts, struct dr_node *all_nodes)
 	if (node == NULL)
 		return NULL;
 
-	dbg("found node: drc name=%s, index=0x%x, path=%s\n",
+	say(DEBUG, "found node: drc name=%s, index=0x%x, path=%s\n",
 	     node->drc_name, node->drc_index, node->ofdt_path);
 
 	/* Prompt user only if not in noprompt mode */
@@ -545,23 +606,48 @@ remove_work(struct options *opts, struct dr_node *all_nodes)
 	/* Make sure there's something there to remove. */
 	if (node->children == NULL) {
 		process_led(node, LED_OFF);
-		err_msg("There is no configured card to remove from the "
-			"specified PCI slot.\n");
+		say(ERROR, "There is no configured card to remove from the "
+		    "specified PCI slot.\n");
 		return NULL;
 	}
 
-	/* Make sure all the devices are
-	 * not configured before proceeding
-	 */
-	if (get_hp_adapter_status(node->drc_name) == CONFIG) {
-		dbg("unconfiguring adapter in slot[%s]\n", node->drc_name);
-		set_hp_adapter_status(PHP_UNCONFIG_ADAPTER, node->drc_name);
+	if (!opts->pci_virtio) {
+		/* Make sure all the devices are
+		 * not configured before proceeding
+		 */
+		if (get_hp_adapter_status(node->drc_name) == CONFIG) {
+			say(DEBUG, "unconfiguring adapter in slot[%s]\n",
+				node->drc_name);
+			set_hp_adapter_status(PHP_UNCONFIG_ADAPTER, node->drc_name);
 
-		int rc = get_hp_adapter_status(node->drc_name);
-		if (rc != NOT_CONFIG) {
-			err_msg("Unconfig adapter failed.\n");
-			return NULL;
+			int rc = get_hp_adapter_status(node->drc_name);
+			if (rc != NOT_CONFIG) {
+				say(ERROR, "Unconfig adapter failed.\n");
+				return NULL;
+			}
+		} else {
+			/* In certain cases such as a complete failure of the
+			 * adapter there may not have been the possibility to clean
+			 * up everything. Mark these adapaters for additional
+			 * processing later.
+			 */
+			node->post_replace_processing = 1;
 		}
+	} else {
+		/* The rpaphp module fails to work with qemu pci slots so
+		 * so we rely on the devices remove attribute in sysfs to
+		 * inform the generic kernel pci code to remove the qemu virtio
+		 * device
+		 */
+		for (child = node->children; child; child = child->next)
+			pci_remove_device(child);
+
+		/* The kernel adds the pci device removal to a work queue which
+		 * makes the previous pci_remove_device call asynchronous. Need
+		 * to let kernel finish cleanup before informing qemu to isolate
+		 * and power down virtio pci device.
+		 */
+		sleep(3);
 	}
 
 	/* Call subroutine to remove node(s) from
@@ -570,7 +656,7 @@ remove_work(struct options *opts, struct dr_node *all_nodes)
 	for (child = node->children; child; child = child->next) {
 		rc = remove_device_tree_nodes(child->ofdt_path);
 		if (rc) {
-			err_msg("%s", sw_error);
+			say(ERROR, "%s", sw_error);
 			rtas_set_indicator(ISOLATION_STATE, node->drc_index,
 					   ISOLATE);
 			set_power(node->drc_power, POWER_OFF);
@@ -582,29 +668,29 @@ remove_work(struct options *opts, struct dr_node *all_nodes)
 	 * allowing the user to physically remove
 	 * the card.
 	 */
-	dbg("is calling rtas_set_indicator(ISOLATE index 0x%x)\n",
+	say(DEBUG, "is calling rtas_set_indicator(ISOLATE index 0x%x)\n",
 	    node->drc_index);
 
 	rc = rtas_set_indicator(ISOLATION_STATE, node->drc_index, ISOLATE);
 	if (rc) {
 		if (rc == HW_ERROR)
-			err_msg("%s", hw_error);
+			say(ERROR, "%s", hw_error);
 		else
-			err_msg("%s", sw_error);
+			say(ERROR, "%s", sw_error);
 
 		set_power(node->drc_power, POWER_OFF);
 		return NULL;
 	}
 
-	dbg("is calling set_power(POWER_OFF index 0x%x, power_domain 0x%x)\n",
-	    node->drc_index, node->drc_power);
+	say(DEBUG, "is calling set_power(POWER_OFF index 0x%x, power_domain "
+	    "0x%x)\n", node->drc_index, node->drc_power);
 
 	rc = set_power(node->drc_power, POWER_OFF);
 	if (rc) {
 		if (rc == HW_ERROR)
-			err_msg("%s", hw_error);
+			say(ERROR, "%s", hw_error);
 		else
-			err_msg("%s", sw_error);
+			say(ERROR, "%s", sw_error);
 
 		set_power(node->drc_power, POWER_OFF);
 		return NULL;
@@ -694,11 +780,11 @@ do_replace(struct options *opts, struct dr_node *all_nodes)
 		return -1;
 
 	if (!repl_node->children) {
-		err_msg("Bad node struct.\n");
+		say(ERROR, "Bad node struct.\n");
 		return -1;
 	}
 
-	dbg("repl_node:path=%s node:path=%s\n",
+	say(DEBUG, "repl_node:path=%s node:path=%s\n",
 	    repl_node->ofdt_path, repl_node->children->ofdt_path);
 
 	/* Prompt user to replace card and to press
@@ -721,28 +807,50 @@ do_replace(struct options *opts, struct dr_node *all_nodes)
 		}
 	}
 
-	rc = add_work(repl_node);
+	rc = add_work(opts, repl_node);
 	if (rc)
 		return rc;
 
-	dbg("CONFIGURING the card in node[name=%s, path=%s]\n",
+	say(DEBUG, "CONFIGURING the card in node[name=%s, path=%s]\n",
 	    repl_node->drc_name, repl_node->ofdt_path);
 
 	set_hp_adapter_status(PHP_CONFIG_ADAPTER, repl_node->drc_name);
 
-	return 0;
+	if (repl_node->post_replace_processing) {
+		int prompt_save = opts->noprompt;
+
+		say(DEBUG, "Doing post replacement processing...\n");
+		/* disable prompting for post-processing */
+		opts->noprompt = 1;
+
+		repl_node = remove_work(opts, repl_node);
+		rc = add_work(opts, repl_node);
+		if (!rc)
+			set_hp_adapter_status(PHP_CONFIG_ADAPTER,
+					      repl_node->drc_name);
+
+		opts->noprompt = prompt_save;
+	}
+
+	return rc;
 }
 
 int
 valid_pci_options(struct options *opts)
 {
 	if ((opts->action == IDENTIFY) && (opts->no_ident)) {
-		err_msg("Cannot specify the -i and -I option together\n");
+		say(ERROR, "Cannot specify the -i and -I option together\n");
 		return -1;
 	}
 
-	if (opts->usr_drc_name == NULL) {
-		err_msg("A drc name must be specified\n");
+	/* The -s option can specify a drc name or drc index */
+	if (!strncmp(opts->usr_drc_name, "0x", 2)) {
+		opts->usr_drc_index = strtoul(opts->usr_drc_name, NULL, 16);
+		opts->usr_drc_name = NULL;
+	}
+
+	if (opts->usr_drc_name == NULL && !opts->usr_drc_index) {
+		say(ERROR, "A drc name or index must be specified\n");
 		return -1;
 	}
 
@@ -757,7 +865,7 @@ drslot_chrp_pci(struct options *opts)
 
 	all_nodes = get_hp_nodes();
 	if (all_nodes == NULL) {
-		err_msg("There are no PCI hot plug slots on this system.\n");
+		say(ERROR, "There are no PCI hot plug slots on this system.\n");
 		return -1;
 	}
 
@@ -765,6 +873,9 @@ drslot_chrp_pci(struct options *opts)
 	print_slots_list(all_nodes);
 #endif
 
+	if (!opts->usr_drc_name)
+		opts->usr_drc_name = find_drc_name(opts->usr_drc_index, all_nodes);
+
 	switch (opts->action) {
 	    case ADD:
 		rc = do_add(opts, all_nodes);
@@ -779,7 +890,7 @@ drslot_chrp_pci(struct options *opts)
 		rc = do_identify(opts, all_nodes);
 		break;
 	    default:
-		err_msg("Invalid operation specified!\n");
+		say(ERROR, "Invalid operation specified!\n");
 		rc = -1;
 		break;
 	}
