@@ -14,10 +14,10 @@
 #include "dr.h"
 #include "lsslot.h"
 #include "drmem.h"
+#include "pseries_platform.h"
 
-int debug = 0;
+int output_level = 0;
 int log_fd = 0;
-char *cmd = "lsslot";
 
 /**
  * struct print_node
@@ -49,9 +49,11 @@ usage(void)
 	fprintf(stderr, "                Display the slots of the specified "
 		"connector type.  The valid\n");
 	fprintf(stderr, "                connector types are \"pci\" for "
-		"hotplg PCI slots, \"slot\" for\n");
-	fprintf(stderr, "                logical slots, \"phb\" fot PHB's "
-		"and \"cpu\" for cpu's. The default\n");
+		"hotplug PCI slots, \"slot\" for\n");
+	fprintf(stderr, "                logical slots, \"phb\" for PHB's, "
+		"\"port\" for LHEA ports, \"mem\"\n");
+	fprintf(stderr, "                for memory, and \"cpu\" "
+		"for cpu's. The default\n");
 	fprintf(stderr, "                is \"slot\" if no -c option is "
 		"specified.\n");
 	fprintf(stderr, "        -a      Display available slots, valid for "
@@ -420,7 +422,7 @@ parse_options(int argc, char *argv[], struct cmd_opts *opts)
 			break;
 
 		    case 'd':
-			set_debug(atoi(optarg));
+			set_output_level(atoi(optarg));
 			break;
 
 		    case 'F':
@@ -431,9 +433,9 @@ parse_options(int argc, char *argv[], struct cmd_opts *opts)
 			 */
 			if ((opts->delim[1] != '\0')
 			    || (opts->delim[0] == '%')) {
-				err_msg("You may specify only one character "
-					"for the -F option,\nand it must not "
-					"be the %% character.\n");
+				say(ERROR, "You may specify only one character "
+				    "for the -F option,\nand it must not "
+				    "be the %% character.\n");
 				exit(1);
 			}
 			break;
@@ -512,8 +514,8 @@ parse_options(int argc, char *argv[], struct cmd_opts *opts)
 			usage();
 
 		if (opts->b_flag && opts->p_flag) {
-			err_msg("You cannot specify both the -b and -p "
-				"options.\n");
+			say(ERROR, "You cannot specify both the -b and -p "
+			    "options.\n");
 			usage();
 		}
 
@@ -554,11 +556,11 @@ lsslot_chrp_pci(struct cmd_opts *opts)
 	/* If nothing returned, then no hot plug node */
 	if (all_nodes == NULL) {
 		if (opts->slot_type == PCI)
-			err_msg("There are no PCI hot plug slots on "
-				"this system.\n");
+			say(ERROR, "There are no PCI hot plug slots on "
+			    "this system.\n");
 		else
-			err_msg("There are no DR slots on this system.\n");
-   		return 1;
+			say(ERROR, "There are no DR slots on this system.\n");
+   		return 0;
 	}
 
 	print_node_list(all_nodes);
@@ -589,8 +591,8 @@ lsslot_chrp_pci(struct cmd_opts *opts)
 		 * user specified a slot or a device name.
 		 */
 		if (opts->s_name != NULL) {
-			err_msg("The specified PCI slot is either invalid\n"
-				"or does not support hot plug operations.\n");
+			say(ERROR, "The specified PCI slot is either invalid\n"
+			    "or does not support hot plug operations.\n");
 			rc = 1;
 		}
 		goto lsslot_pci_exit;
@@ -778,7 +780,7 @@ lsslot_chrp_port(struct cmd_opts *opts)
 
 	/* If nothing returned, then no hot plug node */
 	if (all_nodes == NULL) {
-		err_msg("There are no LHEA ports on this system.\n");
+		say(ERROR, "There are no LHEA ports on this system.\n");
 		return 1;
 	}
 
@@ -810,7 +812,7 @@ lsslot_chrp_port(struct cmd_opts *opts)
 		 * user specified a slot or a device name.
 		 */
 		if (opts->s_name != NULL) {
-			err_msg("The specified port was not found.\n");
+			say(ERROR, "The specified port was not found.\n");
 			rc = 1;
 		}
 		goto lsslot_port_exit;
@@ -846,6 +848,14 @@ main(int argc, char *argv[])
 	struct cmd_opts opts;
 	int rc;
 
+	switch (get_platform()) {
+	case PLATFORM_UNKNOWN:
+	case PLATFORM_POWERKVM_HOST:
+		fprintf(stderr, "%s: is not supported on the %s platform\n",
+						argv[0], platform_name);
+		exit(1);
+	}
+
 	/* make sure that we're running on the proper platform.	*/
 	if (! valid_platform("chrp"))
 		exit(1);
@@ -856,10 +866,10 @@ main(int argc, char *argv[])
 	opts.slot_type = SLOT;
 	parse_options(argc, argv, &opts);
 
-	rc = dr_lock(opts.timeout);
+	rc = dr_lock();
 	if (rc) {
-		err_msg("Unable to obtain Dynamic Reconfiguration lock. "
-			"Please try command again later.\n");
+		say(ERROR, "Unable to obtain Dynamic Reconfiguration lock. "
+		    "Please try command again later.\n");
 		exit(1);
 	}
 
