@@ -21,7 +21,7 @@
  * following NVRAM partitions contain data formatted as name=value pairs: 
  * common, of-config, and ibm,setupcfg.
  *
- * @author Nathan Fontenot <nfont@austin.ibm.com>
+ * @author Nathan Fontenot <nfont@linux.vnet.ibm.com>
  * @author Michael Strosaker <strosake@us.ibm.com>
  * @author Todd Inglett <tinglett@us.ibm.com>
  */
@@ -44,6 +44,7 @@
 #include <getopt.h>
 #include <inttypes.h>
 #include <zlib.h>
+#include <endian.h>
 
 #include "nvram.h"
 
@@ -57,6 +58,7 @@ static int verbose;
 static struct option long_options[] = {
     {"verbose", 		optional_argument, NULL, 'v'},
     {"print-config",		optional_argument, NULL, 'o'},
+    {"zero",			optional_argument, NULL, '0'},
     {"print-vpd", 		optional_argument, NULL, 'V'},
     {"print-all-vpd", 		optional_argument, NULL, 'W'},
     {"print-err-log", 		no_argument, 	   NULL, 'e'},
@@ -84,6 +86,8 @@ help(void)
     "  --print-config[=var]\n"
     "          print value of a config variable, or print all variables in\n"
     "          the specified (or all) partitions\n"
+    "  --zero | -0\n"
+    "          terminate config pairs with a NUL character\n"
     "  --update-config <var>=<value>\n"
     "          update the config variable in the specified partition; the -p\n"
     "          option must also be specified\n"
@@ -191,199 +195,6 @@ warn_msg(const char *fmt, ...)
     va_end(ap);
 }
 
-/**
- * resolve_of_node
- * @brief resolve an Open Firmware node name
- *
- * In a device tree node with a single child, "foo@0", all of the following
- * names refer to that child: "foo@0", "foo", "@0".
- *
- * @param parent fully-qualified path to the parent device node
- * @param node the (possibly abbreviated) name of the child node
- * @param nodelen length of the node parameter
- * @param resolved pointer to return resolved node name in
- * @return 0 if a suitable child can be found, and 'resolved' is a
- * dynamically-allocated string containing the resolved node name.
- * Otherwise, -1.
- */
-static int
-resolve_of_node(const char *parent, const char *node, int nodelen,
-	    char **resolved)
-{
-    static char nodebuf[1024]; /* XXX hardcoded length */
-    struct stat sbuf;
-    glob_t pglob;
-    int rc;
-
-    *resolved = NULL;
-
-    snprintf(nodebuf, sizeof(nodebuf), "%s/%.*s", parent,
-	    nodelen, node);
-    rc = stat(nodebuf, &sbuf);
-    if (rc != -1) {
-	*resolved = malloc(nodelen+2);
-	snprintf(*resolved, nodelen+2, "/%.*s", nodelen, node);
-	goto out;
-    }
-    if (errno != ENOENT)
-	goto out; /* report unusual errors */
-
-    if (node[0] == '@') {
-	/* it's a unit address; glob for *@unitaddr */
-	snprintf(nodebuf, sizeof(nodebuf), "%s/*%.*s*", parent, nodelen, node);
-	rc = glob(nodebuf, 0, NULL, &pglob);
-	if (rc == 0) {
-	    if (pglob.gl_pathc > 1) {
-		err_msg("Ambiguous node name \"%.*s\"\n", nodelen, node);
-		while (pglob.gl_pathc) {
-		    free(pglob.gl_pathv[pglob.gl_pathc]);
-		    pglob.gl_pathc--;
-		}
-		goto out;
-	    }
-
-	    /* skip the leading fully-qualified path */
-	    *resolved = strdup(pglob.gl_pathv[0] + strlen(parent));
-	    free(pglob.gl_pathv[0]);
-	    goto out;
-	}
-    } else {
-	/* must be a node name; glob for node@* */
-	snprintf(nodebuf, sizeof(nodebuf), "%s/%.*s@*", parent, nodelen, node);
-	rc = glob(nodebuf, 0, NULL, &pglob);
-	if (rc == 0) {
-	    if (pglob.gl_pathc > 1) {
-		err_msg("Ambiguous node name \"%.*s\"\n", nodelen, node);
-		while (pglob.gl_pathc) {
-		    free(pglob.gl_pathv[pglob.gl_pathc]);
-		    pglob.gl_pathc--;
-		}
-		goto out;
-	    }
-
-	    /* skip the leading fully-qualified path */
-	    *resolved = strdup(pglob.gl_pathv[0] + strlen(parent));
-	    free(pglob.gl_pathv[0]);
-	    goto out;
-	}
-    }
-
-out:
-    if (*resolved != NULL)
-	return 0;
-    return -1;
-}
-
-/**
- * open_of_path
- * @brief open an Open Firmware path under DEVICE_TREE
- *
- * @param ofpath the path to open, such as "/pci/@d/mac-io/nvram/#bytes"
- *
- * An Open Firmware path may contain "shortcut" node names that are not present
- * under /proc/device-tree. In the above example, we may need to open
- * "pci@80000000" instead of "pci".
- *
- * @return file descriptor to the opened node, or -1 on failure
- */
-static int
-open_of_path(const char *ofpath)
-{
-    static char resolved_ofpath[1024]; /* XXX hardcoded length */
-    const char *node;
-    int fd = -1;
-    int rc = 0;
-
-    strcpy(resolved_ofpath, DEVICE_TREE);
-
-    while (ofpath) {
-	int nodelen;
-	char *resolved_node;
-
-	node = ofpath + 1;
-	ofpath = strchr(node + 1, '/');
-
-	nodelen = ofpath - node;
-	if (ofpath == NULL)
-	    nodelen = strlen(node);
-
-	rc = resolve_of_node(resolved_ofpath, node, nodelen, &resolved_node);
-	if (rc < 0)
-	    break;
-
-	strcat(resolved_ofpath, resolved_node);
-	free(resolved_node);
-    }
-
-    if (rc >= 0)
-	fd = open(resolved_ofpath, O_RDONLY);
-
-    return fd;
-}
-
-/**
- * get_of_nvram_size
- * @brief Get the size of nvram from the device tree
- *
- * Retrieve the size of nvram as specified by the Open Firmware
- * device tree.  If this fails we return a default size of
- * 1024 * 1024.
- *
- * @return size of nvram
- */
-static int 
-get_of_nvram_size(void)
-{
-    char buf[1024] = NVRAM_DEFAULT "/#bytes";
-    int fd;
-    int size, len;
-
-    fd = open(buf, O_RDONLY);
-    if (fd == -1) {
-	/* Check the aliases directory */
-	struct stat sbuf;
-	int offset;
-
-        fd = open(NVRAM_ALIAS, O_RDONLY);
-	if (fd == -1) {
-	    err_msg("%s", "Could not determine nvram size from "
-		    NVRAM_ALIAS "\n");
-	    return DEFAULT_NVRAM_SZ;
-	}
-
-	if (fstat(fd, &sbuf) != 0) {
-	    err_msg("%s", "Could not determine nvram size from "
-		    NVRAM_ALIAS "\n");
-	    close(fd);
-	    return DEFAULT_NVRAM_SZ;
-	}
-
-	offset = read(fd, buf, sbuf.st_size - 1);
-	offset += sprintf(buf + offset, "%s", "/#bytes");
-	buf[offset] = '\0';
-
-	close(fd);
-	fd = open_of_path(buf);
-    }
-
-    if (fd == -1) {
-	warn_msg("cannot open nvram node \"%s\" in device tree: %s\n", buf,
-		strerror(errno));
-	close(fd);
-	return DEFAULT_NVRAM_SZ;
-    }
-    
-    len = read(fd, &size, sizeof(size));
-    close(fd);
-    
-    if (len != sizeof(size)) {
-	perror("got odd size for nvram node in device tree");
-	return DEFAULT_NVRAM_SZ;
-    }
-    
-    return size;
-}
-
 /**
  * nvram_read
  * @brief read in the contents of nvram
@@ -629,6 +440,7 @@ nvram_parse_partitions(struct nvram *nvram)
 	c_sum = checksum(phead);
 	if (c_sum != phead->checksum)
 	    warn_msg("this partition checksum should be %02x!\n", c_sum);
+	phead->length = be16toh(phead->length);
 	p_start += phead->length * NVRAM_BLOCK_SIZE;
     }
 
@@ -674,7 +486,8 @@ nvram_find_fd_partition(struct nvram *nvram, char *name)
 	if (! strncmp(phead.name, name, sizeof(phead.name)))
 	    found = 1;
 	else {
-	    int offset = phead.length * NVRAM_BLOCK_SIZE - len;
+	    int offset =
+		    be16toh(phead.length) * NVRAM_BLOCK_SIZE - len;
 	    if (lseek(nvram->fd, offset, SEEK_CUR) == -1) {
 	        err_msg("seek error in file %s: %s\n", nvram->filename,
 			strerror(errno));
@@ -1299,7 +1112,16 @@ unzip_partition(struct nvram *nvram, char *name)
     next = start + sizeof(*phead);	/* Skip partition header. */
     next += sizeof(struct err_log_info);	/* Skip sub-header. */
     zipped_length = *((unsigned short*) next);
-    next += sizeof(unsigned short);		/* Skip compressed length. */
+    next += sizeof(unsigned short);
+
+   /* New format oops header, zipped_length > OOPS_PARTITION_SZ
+    * signifies the version of new header. Find out new zipped length
+    * and from where the compressed data starts.
+    */
+   if (zipped_length > OOPS_PARTITION_SZ) {
+        zipped_length = *((unsigned short*) next);
+        next += sizeof(struct oops_log_info) - sizeof(unsigned short);
+   }
 
     if ((next-start) + zipped_length > phead->length * NVRAM_BLOCK_SIZE) {
     	err_msg("bogus size for compressed data in partition %s: %u\n", name,
@@ -1374,13 +1196,18 @@ static int num_name_value_parts = 3;
  * @return 0 on success, !0 otherwise
  */
 static int 
-print_of_config(struct nvram *nvram, char *config_var, char *pname)
+print_of_config(struct nvram *nvram, char *config_var, char *pname,
+	int zero_terminator)
 {
     struct partition_header *phead;
-    char *data;
+    char *data, terminator;
     int  i, varlen;
     int  rc = -1;
 
+    terminator = '\n';
+    if (zero_terminator)
+	terminator = '\0';
+
     /* if config_var is NULL , print the data from the
      * partition specified by pname or all of the
      * name/value pair partitions if pname is NULL. 
@@ -1417,7 +1244,7 @@ print_of_config(struct nvram *nvram, char *config_var, char *pname)
 	    while (*data != '\0') {
 	    	if ((data[varlen] == '=') && 
 		    strncmp(config_var, data, varlen) == 0) {
-    	            printf("%s\n", data + varlen + 1);
+    	            printf("%s%c", data + varlen + 1, terminator);
 		    rc = 0;
 		}
 	    	data += strlen(data) + 1;
@@ -1435,7 +1262,7 @@ print_of_config(struct nvram *nvram, char *config_var, char *pname)
 	while (*data != '\0') {
 	    if ((data[varlen] == '=') && 
 		strncmp(config_var, data, varlen) == 0) {
-		printf("%s\n", data + varlen + 1);
+		printf("%s%c", data + varlen + 1, terminator);
 		rc = 0;
 	    }
 	    data += strlen(data) + 1;
@@ -1463,12 +1290,20 @@ int
 update_of_config_var(struct nvram *nvram, char *config_var, char *pname)
 {
     struct partition_header *phead, *new_phead;
+    char *new_config_value;
     char *data_offset;
     char *new_part;
     char *new_part_offset, *new_part_end;
     char *tmp_offset;
     int	config_name_len;
-    int	len, part_size, found = 0;
+    int	len, rc, part_size;
+
+    new_config_value = strchr(config_var, '=');
+    if (!new_config_value) {
+	err_msg("config variables must be in the format \"name=value\"");
+	return -1;
+    }
+    new_config_value++;
 
     phead = nvram_find_partition(nvram, 0, pname, NULL);
     if (phead == NULL) {
@@ -1493,20 +1328,9 @@ update_of_config_var(struct nvram *nvram, char *config_var, char *pname)
     
     /* now find this config variable in the partition */
     while (*data_offset != '\0') {
-	if (strncmp(data_offset, config_var, config_name_len) == 0) {
-            found = 1;
+	if (strncmp(data_offset, config_var, config_name_len) == 0)
 	    break;
-        }
-	else
-	    data_offset += strlen(data_offset) + 1;
-    }
-
-    if (!found) {
-        err_msg("cannot update %s\n"
-		"\tThe config var does not exist in the \"%s\" partition\n", 
-		config_var, pname);
-	free(new_part);
-	return -1;
+	data_offset += strlen(data_offset) + 1;
     }
 
     /* Copy everything up to the config name we are modifying 
@@ -1516,6 +1340,7 @@ update_of_config_var(struct nvram *nvram, char *config_var, char *pname)
 
     /* make sure the new config var will fit into the partition and add it */
     new_phead = (struct partition_header *)new_part;
+    new_phead->length = be16toh(new_phead->length);
     new_part_offset = new_part + (data_offset - (char *)phead);
     new_part_end = new_part + part_size;
 
@@ -1527,9 +1352,11 @@ update_of_config_var(struct nvram *nvram, char *config_var, char *pname)
 	return -1;
     }
 
-    strncpy(new_part_offset, config_var, strlen(config_var));
-    new_part_offset += strlen(config_var);
-    *new_part_offset++ = '\0';
+    if (strlen(new_config_value)) {
+	strncpy(new_part_offset, config_var, strlen(config_var));
+	new_part_offset += strlen(config_var);
+	*new_part_offset++ = '\0';
+    }
 
     /* Find the end of the name/value pairs in the partition so we
      * can copy them over to the new partition.
@@ -1571,7 +1398,12 @@ update_of_config_var(struct nvram *nvram, char *config_var, char *pname)
     }
     
     /* write the partition out to nvram */
-    len = write(nvram->fd, new_part, part_size);
+    for (rc = 0, len = 0; len < part_size; len += rc) {
+	rc = write(nvram->fd, new_part + len, part_size - len);
+	if (rc <= 0)
+	    break;
+    }
+
     if (len != part_size) {
         err_msg("only wrote %d bytes of the \"%s\" partition back\n"
 		"\tto %s, expected to write %d bytes\n",
@@ -1588,7 +1420,6 @@ main (int argc, char *argv[])
     struct nvram nvram;
     struct stat sbuf;
     int ret = 0; 
-    int	of_nvram_size;
     int	option_index;
     char *endp;
     char *of_config_var = NULL;
@@ -1597,6 +1428,7 @@ main (int argc, char *argv[])
     int print_errlog = 0;
     int print_event_scan = 0;
     int	print_config_var = 0;
+    int zero_terminator = 0;
     char *dump_name = NULL;
     char *ascii_name = NULL;
     char *zip_name = NULL;
@@ -1615,7 +1447,7 @@ main (int argc, char *argv[])
 	
     for (;;) {
 	option_index = 0;
-	ret = getopt_long(argc, argv, "+p:Vv::", long_options, &option_index);
+	ret = getopt_long(argc, argv, "+p:Vv::0", long_options, &option_index);
 	if (ret == -1)
 		break;
 	switch (ret) {
@@ -1641,6 +1473,9 @@ main (int argc, char *argv[])
 		print_config_var = 1;
 		of_config_var = optarg;
 		break;
+	    case '0':
+		zero_terminator = 1;
+		break;
 	    case 'P':	/* partitions */
 		print_partitions = 1;
 		break;
@@ -1720,11 +1555,22 @@ main (int argc, char *argv[])
 	goto err_exit;
     }
 
-    of_nvram_size = get_of_nvram_size();
-    nvram.nbytes = sbuf.st_size ? sbuf.st_size : of_nvram_size;
-    if (nvram.nbytes != of_nvram_size) {
-	warn_msg("specified nvram size %d does not match this machine %d!\n", 
-		 nvram.nbytes, of_nvram_size);
+    if (!nvram.nbytes) {
+	ret = lseek(nvram.fd, 0, SEEK_END);
+	if (ret < 0) {
+	    err_msg("cannot seek(END) %s: %s\n", nvram.filename,
+		    strerror(errno));
+	    goto err_exit;
+	}
+
+	nvram.nbytes = ret;
+
+	ret = lseek(nvram.fd, 0, SEEK_SET);
+	if (ret < 0) {
+	    err_msg("cannot seek(SET) %s: %s\n", nvram.filename,
+		    strerror(errno));
+	    goto err_exit;
+	}
     }
 
     nvram.data = malloc(nvram.nbytes);
@@ -1757,7 +1603,8 @@ main (int argc, char *argv[])
 	    ret = -1; 
     }
     if (print_config_var)
-	if (print_of_config(&nvram, of_config_var, config_pname) != 0)
+	if (print_of_config(&nvram, of_config_var, config_pname,
+		    zero_terminator) != 0)
 	    ret = -1;
     if (print_vpd)
 	if (dump_vpd(&nvram, print_vpd == 2) != 0)
