@@ -14,21 +14,27 @@
 #include <string.h>
 #include <dirent.h>
 #include <librtas.h>
+#include <sys/types.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #include <getopt.h>
 #include <sched.h>
 #include <assert.h>
 #include <pthread.h>
 #include <sys/ioctl.h>
+#include <sys/ptrace.h>
+#include <sys/wait.h>
+#include <sys/resource.h>
 #ifdef HAVE_LINUX_PERF_EVENT_H
 #include <linux/perf_event.h>
 #endif
 #include "librtas_error.h"
 #include <errno.h>
 
-#define PPC64_CPU_VERSION	"1.1"
+#define PPC64_CPU_VERSION	"1.2"
 
 #define SYSFS_CPUDIR	"/sys/devices/system/cpu/cpu%d"
+#define SYSFS_SUBCORES	"/sys/devices/system/cpu/subcores_per_core"
 #define DSCR_DEFAULT_PATH "/sys/devices/system/cpu/dscr_default"
 #define INTSERV_PATH	"/proc/device-tree/cpus/%s/ibm,ppc-interrupt-server#s"
 
@@ -47,11 +53,11 @@ static int counters[MAX_NR_CPUS];
 
 #endif
 
-int threads_per_cpu = 0;
-int cpus_in_system = 0;
-int threads_in_system = 0;
+static int threads_per_cpu = 0;
+static int cpus_in_system = 0;
+static int threads_in_system = 0;
 
-int test_attr(char *path, char *perms)
+static int test_attr(char *path, char *perms)
 {
 	FILE *fp;
 
@@ -61,24 +67,26 @@ int test_attr(char *path, char *perms)
 		return 1;
 	}
 
-	if (errno == ENOENT)
-		/* cpu probably offline */
-		return 1;
+	if (errno == ENOENT) {
+		/* When CPU is offline, attr may not exist. Return 2 so caller
+		 * will not see 'false' and continue to try next CPU. */
+		return 2;
+	}
 
 	return 0;
 }
 
-int attr_is_readable(char *path)
+static int attr_is_readable(char *path)
 {
 	return test_attr(path, "r");
 }
 
-int attr_is_writeable(char *path)
+static int attr_is_writeable(char *path)
 {
 	return test_attr(path, "w");
 }
 
-int test_sysattr(char *attribute, char *perms)
+static int test_sysattr(char *attribute, char *perms)
 {
 	char path[SYSFS_PATH_MAX];
 	int i;
@@ -92,17 +100,17 @@ int test_sysattr(char *attribute, char *perms)
 	return 1;
 }
 
-int sysattr_is_readable(char *attribute)
+static int sysattr_is_readable(char *attribute)
 {
 	return test_sysattr(attribute, "r");
 }
 
-int sysattr_is_writeable(char *attribute)
+static int sysattr_is_writeable(char *attribute)
 {
 	return test_sysattr(attribute, "w");
 }
 
-int get_attribute(char *path, const char *fmt, int *value)
+static int get_attribute(char *path, const char *fmt, int *value)
 {
 	FILE *fp;
 	int rc;
@@ -119,24 +127,39 @@ int get_attribute(char *path, const char *fmt, int *value)
 	rc = fscanf(fp, fmt, value);
 	fclose(fp);
 
-	return 0;
-}
-
-int set_attribute(const char *path, const char *fmt, int value)
-{
-	FILE *fp;
-
-	fp = fopen(path, "w");
-	if (fp == NULL)
+	if (rc == EOF)
 		return -1;
 
-	fprintf(fp, fmt, value);
-	fclose(fp);
-
 	return 0;
 }
 
-int cpu_online(int thread)
+static int set_attribute(const char *path, const char *fmt, int value)
+{
+	int fd, rc, len;
+	char *str;
+
+	fd = open(path, O_WRONLY);
+	if (fd < 0)
+		return -1;
+
+	len = asprintf(&str, fmt, value);
+	if (len < 0) {
+		rc = -1;
+		goto close;
+	}
+
+	rc = write(fd, str, len);
+	free(str);
+
+	if (rc == len)
+		rc = 0;
+
+close:
+	close(fd);
+	return rc;
+}
+
+static int cpu_online(int thread)
 {
 	char path[SYSFS_PATH_MAX];
 	int rc, online;
@@ -149,7 +172,8 @@ int cpu_online(int thread)
 	return 1;
 }
 
-int get_system_attribute(char *attribute, const char *fmt, int *value)
+static int get_system_attribute(char *attribute, const char *fmt, int *value,
+			 int *inconsistent)
 {
 	char path[SYSFS_PATH_MAX];
 	int i, rc;
@@ -165,15 +189,17 @@ int get_system_attribute(char *attribute, const char *fmt, int *value)
 
 		if (system_attribute == -1)
 			system_attribute = cpu_attribute;
-		else if (system_attribute != cpu_attribute)
+		else if (system_attribute != cpu_attribute) {
+			*inconsistent = 1;
 			return -1;
+		}
 	}
 
 	*value = system_attribute;
 	return 0;
 }
 
-int set_system_attribute(char *attribute, const char *fmt, int state)
+static int set_system_attribute(char *attribute, const char *fmt, int state)
 {
 	char path[SYSFS_PATH_MAX];
 	int i, rc;
@@ -181,6 +207,12 @@ int set_system_attribute(char *attribute, const char *fmt, int state)
 	for (i = 0; i < threads_in_system; i++) {
 		sprintf(path, SYSFS_CPUDIR"/%s", i, attribute);
 		rc = set_attribute(path, fmt, state);
+		/* When a CPU is offline some sysfs files are removed from the CPU
+		 * directory, for example smt_snooze_delay and dscr. The absence of the
+		 * file is not an error, so detect and clear the error when
+		 * set_attribute indicates ENOENT. */
+		if (rc == -1 && errno == ENOENT)
+			rc = errno = 0;
 		if (rc)
 			return rc;
 	}
@@ -188,7 +220,7 @@ int set_system_attribute(char *attribute, const char *fmt, int state)
 	return 0;
 }
 
-int dscr_default_exists(void)
+static int dscr_default_exists(void)
 {
 	struct stat sb;
 
@@ -202,7 +234,7 @@ int dscr_default_exists(void)
  * specific dscr value.  This is because the dscr value is now thread
  * specific.
  */
-int set_dscr(int state)
+static int set_dscr(int state)
 {
 	int rc;
 
@@ -225,7 +257,7 @@ int set_dscr(int state)
 	return rc;
 }
 
-int get_dscr(int *value)
+static int get_dscr(int *value, int *inconsistent)
 {
 	int rc;
 
@@ -242,13 +274,13 @@ int get_dscr(int *value)
 			return -2;
 		}
 
-		rc = get_system_attribute("dscr", "%x", value);
+		rc = get_system_attribute("dscr", "%x", value, inconsistent);
 	}
 
 	return rc;
 }
 
-int set_smt_snooze_delay(int delay)
+static int set_smt_snooze_delay(int delay)
 {
 	if (!sysattr_is_writeable("smt_snooze_delay")) {
 		perror("Cannot set smt snooze delay");
@@ -258,33 +290,48 @@ int set_smt_snooze_delay(int delay)
 	return set_system_attribute("smt_snooze_delay", "%d", delay);
 }
 
-int get_smt_snooze_delay(int *delay)
+static int get_smt_snooze_delay(int *delay, int *inconsistent)
 {
 	if (!sysattr_is_readable("smt_snooze_delay")) {
 		perror("Cannot retrieve smt snooze delay");
 		return -2;
 	}
 
-	return get_system_attribute("smt_snooze_delay", "%d", delay);
+	return get_system_attribute("smt_snooze_delay", "%d", delay,
+				    inconsistent);
 }
 
-int online_thread(const char *path)
+static int online_thread(const char *path)
 {
 	return set_attribute(path, "%d", 1);
 }
 
-int offline_thread(const char *path)
+static int offline_thread(const char *path)
 {
 	return set_attribute(path, "%d", 0);
 }
 
+static int is_subcore_capable(void)
+{
+	return attr_is_readable(SYSFS_SUBCORES) == 1;
+}
 
-int get_cpu_info(void)
+static int num_subcores(void)
+{
+	int rc, subcores;
+	rc = get_attribute(SYSFS_SUBCORES, "%d", &subcores);
+	if (rc)
+		return -1;
+	return subcores;
+}
+
+static int get_cpu_info(void)
 {
 	DIR *d;
 	struct dirent *de;
 	int first_cpu = 1;
 	int rc;
+	int subcores;
 
 	d = opendir("/proc/device-tree/cpus");
 	if (!d)
@@ -310,10 +357,16 @@ int get_cpu_info(void)
 
 	closedir(d);
 	threads_in_system = cpus_in_system * threads_per_cpu;
+
+	subcores = num_subcores();
+	if (is_subcore_capable() && subcores > 0) {
+		threads_per_cpu /= subcores;
+		cpus_in_system *= subcores;
+	}
 	return 0;
 }
 
-int is_smt_capable(void)
+static int is_smt_capable(void)
 {
 	struct stat sb;
 	char path[SYSFS_PATH_MAX];
@@ -329,7 +382,7 @@ int is_smt_capable(void)
 	return 0;
 }
 
-int get_one_smt_state(int primary_thread)
+static int get_one_smt_state(int primary_thread)
 {
 	int thread_state;
 	int smt_state = 0;
@@ -343,7 +396,7 @@ int get_one_smt_state(int primary_thread)
 	return smt_state ? smt_state : -1;
 }
 
-int get_smt_state(void)
+static int get_smt_state(void)
 {
 	int system_state = -1;
 	int i;
@@ -369,23 +422,24 @@ int get_smt_state(void)
 	return system_state;
 }
 
-int set_one_smt_state(int thread, int online_threads)
+static int set_one_smt_state(int thread, int online_threads)
 {
 	char path[SYSFS_PATH_MAX];
-	int i, rc;
+	int i, rc = 0;
 
-	for (i = 0; i < online_threads; i++) {
+	for (i = 0; i < threads_per_cpu; i++) {
 		snprintf(path, SYSFS_PATH_MAX, SYSFS_CPUDIR"/%s", thread + i,
 			 "online");
-		rc = online_thread(path);
-		if (rc)
-			return rc;
-	}
+		if (i < online_threads)
+			rc = online_thread(path);
+		else
+			rc = offline_thread(path);
 
-	for (; i < threads_per_cpu; i++) {
-		snprintf(path, SYSFS_PATH_MAX, SYSFS_CPUDIR"/%s", thread + i,
-			 "online");
-		rc = offline_thread(path);
+		/* The 'online' sysfs file returns EINVAL if set to the current
+		 * setting. As this is not an error, reset rc and errno to avoid
+		 * returning failure. */
+		if (rc == -1 && errno == EINVAL)
+			rc = errno = 0;
 		if (rc)
 			break;
 	}
@@ -393,35 +447,48 @@ int set_one_smt_state(int thread, int online_threads)
 	return rc;
 }
 
-int set_smt_state(int smt_state)
+static int set_smt_state(int smt_state)
 {
-	int i, rc;
+	int i, j, rc;
 	int ssd, update_ssd = 1;
+	int inconsistent = 0;
+	int error = 0;
 
 	if (!sysattr_is_writeable("online")) {
 		perror("Cannot set smt state");
 		return -1;
 	}
 
-	rc = get_smt_snooze_delay(&ssd);
+	rc = get_smt_snooze_delay(&ssd, &inconsistent);
 	if (rc)
 		update_ssd = 0;
 
 	for (i = 0; i < threads_in_system; i += threads_per_cpu) {
-		if (cpu_online(i)) {
+		/* Online means any thread on this core running, so check all
+		 * threads in the core, not just the first. */
+		for (j = 0; j < threads_per_cpu; j++) {
+			if (!cpu_online(i + j))
+				continue;
+
 			rc = set_one_smt_state(i, smt_state);
+			/* Record an error, but do not check result: if we
+			 * have failed to set this core, keep trying
+			 * subsequent ones. */
 			if (rc)
-				break;
+				error = 1;
+			break;
 		}
 	}
 
 	if (update_ssd)
 		set_smt_snooze_delay(ssd);
 
+	if (error)
+		return -1;
 	return rc;
 }
 
-int is_dscr_capable(void)
+static int is_dscr_capable(void)
 {
 	struct stat sb;
 	char path[SYSFS_PATH_MAX];
@@ -440,7 +507,7 @@ int is_dscr_capable(void)
 	return 0;
 }
 
-int do_smt(char *state)
+static int do_smt(char *state)
 {
 	int rc = 0;
 	int smt_state;
@@ -483,37 +550,125 @@ int do_smt(char *state)
 	return rc;
 }
 
-int do_dscr(char *state)
+static int do_subcores_per_core(char *state)
 {
 	int rc = 0;
+	int subcore_state = 0;
+
+	/* Check SMT machine. */
+	if (!is_smt_capable()) {
+		fprintf(stderr, "Machine is not SMT capable\n");
+		return -1;
+	}
+
+	/* Check subcore capable machine/kernel. */
+	if (!is_subcore_capable()) {
+		fprintf(stderr, "Machine is not subcore capable\n");
+		return -1;
+	}
+
+	if (!state) {
+		/* Display current status. */
+		subcore_state = num_subcores();
+		if (subcore_state < 0) {
+			fprintf(stderr, "Could not read subcore state.\n");
+			return -1;
+		}
+		printf("Subcores per core: %d\n", subcore_state);
+	} else {
+		subcore_state = strtol(state, NULL, 0);
+		/* Validate option: we only suport 1 or 4. */
+		if (subcore_state != 1 && subcore_state != 4) {
+			fprintf(stderr, "subcores-per-core=%d invalid.\n", subcore_state);
+			return -1;
+		}
+
+		rc = set_attribute(SYSFS_SUBCORES, "%d", subcore_state);
+		if (rc) {
+			fprintf(stderr, "Failed to set subcore option.\n");
+			return rc;
+		}
+
+		printf("Subcores per core set to %d\n", subcore_state);
+	}
+
+	return rc;
+}
+
+#define PTRACE_DSCR 44
+
+static int do_dscr_pid(int dscr_state, pid_t pid)
+{
+	int rc;
+
+	rc = ptrace(PTRACE_ATTACH, pid, NULL, NULL);
+	if (rc) {
+		fprintf(stderr, "Could not attach to process %d to %s the "
+			"DSCR value\n%s\n", pid, (dscr_state ? "set" : "get"),
+			strerror(errno));
+		return rc;
+	}
+
+	wait(NULL);
+
+	if (dscr_state) {
+		rc = ptrace(PTRACE_POKEUSER, pid, PTRACE_DSCR << 3, dscr_state);
+		if (rc) {
+			fprintf(stderr, "Could not set the DSCR value for pid "
+				"%d\n%s\n", pid, strerror(errno));
+			ptrace(PTRACE_DETACH, pid, NULL, NULL);
+			return rc;
+		}
+	}
+
+	rc = ptrace(PTRACE_PEEKUSER, pid, PTRACE_DSCR << 3, NULL);
+	if (errno) {
+		fprintf(stderr, "Could not get the DSCR value for pid "
+			"%d\n%s\n", pid, strerror(errno));
+		rc = -1;
+	} else {
+		printf("DSCR for pid %d is %d\n", pid, rc);
+	}
+
+	ptrace(PTRACE_DETACH, pid, NULL, NULL);
+	return rc;
+}
+
+static int do_dscr(char *state, pid_t pid)
+{
+	int rc = 0;
+	int dscr_state = 0;
 
 	if (!is_dscr_capable()) {
 		fprintf(stderr, "Machine is not DSCR capable\n");
 		return -1;
 	}
 
-	if (!state) {
-		int dscr;
+	if (state)
+		dscr_state = strtol(state, NULL, 0);
 
-		rc = get_dscr(&dscr);
-		switch (rc) {
-		    case -1:
-			printf("Could not retrieve DSCR\n");
-			break;
-		    case 0:
-			if (dscr == -1)
+	if (pid != -1)
+		return do_dscr_pid(dscr_state, pid);
+
+	if (!state) {
+		int dscr, inconsistent = 0;
+
+		rc = get_dscr(&dscr, &inconsistent);
+		if (rc) {
+			if (inconsistent)
 				printf("Inconsistent DSCR\n");
 			else
-				printf("dscr is %d\n", dscr);
-			break;
+				printf("Could not retrieve DSCR\n");
+		} else {
+			printf("DSCR is %d\n", dscr);
 		}
 	} else
-		rc = set_dscr(strtol(state, NULL, 0));
+		rc = set_dscr(dscr_state);
 
 	return rc;
 }
 
-int do_smt_snooze_delay(char *state)
+static int do_smt_snooze_delay(char *state)
 {
 	int rc = 0;
 
@@ -523,15 +678,15 @@ int do_smt_snooze_delay(char *state)
 	}
 
 	if (!state) {
-		int ssd;
-		rc = get_smt_snooze_delay(&ssd);
+		int ssd, inconsistent = 0;
+		rc = get_smt_snooze_delay(&ssd, &inconsistent);
 		if (rc) {
-			printf("Could not retrieve smt_snooze_delay\n");
-		} else {
-			if (ssd == -1)
+			if (inconsistent)
 				printf("Inconsistent smt_snooze_delay\n");
 			else
-				printf("smt_snooze_delay is %d\n", ssd);
+				printf("Could not retrieve smt_snooze_delay\n");
+		} else {
+			printf("smt_snooze_delay is %d\n", ssd);
 		}
 	} else {
 		int delay;
@@ -547,12 +702,12 @@ int do_smt_snooze_delay(char *state)
 	return rc;
 }
 
-int do_run_mode(char *run_mode)
+static int do_run_mode(char *run_mode)
 {
 	char mode[3];
 	int rc;
 
-	if (getpid() != 0) {
+	if (getuid() != 0) {
 		fprintf(stderr, "Cannot %s run mode: Permission denied\n",
 			run_mode ? "set" : "get");
 		return -1;
@@ -702,7 +857,92 @@ static void *soak(void *arg)
 		; /* Do Nothing */
 }
 
-int do_cpu_frequency(void)
+static char *power_mode(uint64_t mode)
+{
+	switch (mode) {
+	case 0x0001:
+		return "Dynamic, Favor Performance\n";
+	case 0x0002:
+		return "None\n";
+	case 0x0003:
+		return "Static\n";
+	case 0x00ff:
+		return "Dynamic, Favor Power\n";
+	default:
+		return "Unknown";
+	}
+}
+
+static void report_system_power_mode(void)
+{
+	FILE *f;
+	char line[128];
+
+	f = fopen("/proc/ppc64/lparcfg", "r");
+	if (!f)
+		return;
+
+	while (fgets(line, 128, f) != NULL) {
+		char *name, *value;
+		uint64_t mode, system_mode, partition_mode;
+
+		if ((line[0] == '\n') || (!strncmp(&line[0], "lparcfg", 7)))
+			continue;
+
+		name = &line[0];
+		value = strchr(line, '=');
+		*value = '\0';
+		value++;
+
+		if (strcmp(name, "power_mode_data"))
+			continue;
+
+		/* The power mode result is defined as
+		 * XXXX XXXX XXXX XXXX
+		 * XXXX			: System Power Mode
+		 *                XXXX	: Partition Power Mode
+		 * They mode is the first 4 bytes of the value reported in
+		 * the lparcfg file.
+		 */
+		mode = strtoul(value, NULL, 16);
+		system_mode = (mode >> 48) & 0xffff;
+		partition_mode = mode & 0xffff;
+
+		if (system_mode != partition_mode) {
+			printf("System Power Savings Mode: %s",
+			       power_mode(system_mode));
+			printf("Partition Power Savings Mode: %s",
+			       power_mode(partition_mode));
+		} else {
+			printf("Power Savings Mode: %s",
+			       power_mode(system_mode));
+		}
+	}
+
+	fclose(f);
+	return;
+}
+
+/* We need an FD per CPU, with a few more for stdin/out/err etc */
+static void setrlimit_open_files(void)
+{
+	struct rlimit old_rlim, new_rlim;
+	int new = threads_in_system + 8;
+
+	getrlimit(RLIMIT_NOFILE, &old_rlim);
+
+	if (old_rlim.rlim_cur > new)
+		return;
+
+	new_rlim.rlim_cur = new;
+	new_rlim.rlim_max = old_rlim.rlim_max;
+
+	setrlimit(RLIMIT_NOFILE, &new_rlim);
+}
+
+#define freq_calc(cycles, time)	(1.0 * (cycles) / (time) / 1000000000ULL)
+
+static int do_cpu_frequency(int sleep_time)
 {
 	int i, rc;
 	unsigned long long min = -1ULL;
@@ -712,6 +952,8 @@ int do_cpu_frequency(void)
 	unsigned long long sum = 0;
 	unsigned long count = 0;
 
+	setrlimit_open_files();
+
 	memset(cpu_freq, 0, sizeof(cpu_freq));
 	memset(counters, 0, sizeof(counters));
 
@@ -738,8 +980,8 @@ int do_cpu_frequency(void)
 	usleep(1000000);
 
 	start_counters();
-	/* Count for 1 second */
-	usleep(1000000);
+	/* Count for specified timeout in seconds */
+	usleep(sleep_time * 1000000);
 	stop_counters();
 
 	read_counters();
@@ -766,17 +1008,18 @@ int do_cpu_frequency(void)
 		count++;
 	}
 
-	printf("min:\t%.2f GHz (cpu %ld)\n", 1.0 * min / 1000000000ULL,
+	report_system_power_mode();
+	printf("min:\t%.3f GHz (cpu %ld)\n", freq_calc(min, sleep_time),
 	       min_cpu);
-	printf("max:\t%.2f GHz (cpu %ld)\n", 1.0 * max / 1000000000ULL,
+	printf("max:\t%.3f GHz (cpu %ld)\n", freq_calc(max, sleep_time),
 	       max_cpu);
-	printf("avg:\t%.2f GHz\n\n", 1.0 * (sum / count) / 1000000000ULL);
+	printf("avg:\t%.3f GHz\n\n", freq_calc((sum / count), sleep_time));
 	return 0;
 }
 
 #else
 
-int do_cpu_frequency(void)
+static int do_cpu_frequency(void)
 {
 	printf("CPU Frequency determination is not supported on this "
 	       "platfom.\n");
@@ -785,13 +1028,13 @@ int do_cpu_frequency(void)
 
 #endif
 
-int do_cores_present(char * state)
+static int do_cores_present(char * state)
 {
 	printf("Number of cores present = %d\n", cpus_in_system);
 	return 0;
 }
 
-int set_all_threads_off(int cpu, int smt_state)
+static int set_all_threads_off(int cpu, int smt_state)
 {
 	int i;
 	char path[SYSFS_PATH_MAX];
@@ -807,7 +1050,7 @@ int set_all_threads_off(int cpu, int smt_state)
 	return rc;
 }
 
-int set_one_core(int smt_state, int core, int state)
+static int set_one_core(int smt_state, int core, int state)
 {
 	int rc = 0;
 	int cpu = core * threads_per_cpu;
@@ -825,7 +1068,7 @@ int set_one_core(int smt_state, int core, int state)
 	return rc;
 }
 
-int do_cores_online(char *state)
+static int do_cores_online(char *state)
 {
 	int smt_state;
 	int *core_state;
@@ -869,8 +1112,11 @@ int do_cores_online(char *state)
 	if (number_to_have == cores_now_online)
 		return 0;
 
-	if (number_to_have > cpus_in_system)
-		number_to_have = cpus_in_system;
+	if (number_to_have > cpus_in_system) {
+		printf("Cannot online more cores than are present.\n");
+		do_cores_present(NULL);
+		return -1;
+	}
 
 	if (number_to_have > cores_now_online) {
 		number_to_change = number_to_have - cores_now_online;
@@ -903,22 +1149,58 @@ int do_cores_online(char *state)
 	return 0;
 }
 
-void usage(void)
+static int do_info(char *arg)
 {
-	printf("\tppc64_cpu --smt               # Get current SMT state\n"
-	       "\tppc64_cpu --smt={on|off}      # Turn SMT on/off\n"
-	       "\tppc64_cpu --smt=X             # Set SMT state to X\n\n"
-	       "\tppc64_cpu --cores-present     # Get the number of cores installed\n"
-	       "\tppc64_cpu --cores-on          # Get the number of cores currently online\n"
-	       "\tppc64_cpu --cores-on=X        # Put exactly X cores online\n\n"
+	int i, j, thread_num;
+	char online;
+	int subcores = 0;
 
-	       "\tppc64_cpu --dscr              # Get current DSCR setting\n"
-	       "\tppc64_cpu --dscr=<val>        # Change DSCR setting\n\n"
-	       "\tppc64_cpu --smt-snooze-delay  # Get current smt-snooze-delay setting\n"
-	       "\tppc64_cpu --smt-snooze-delay=<val> # Change smt-snooze-delay setting\n\n"
-	       "\tppc64_cpu --run-mode          # Get current diagnostics run mode\n"
-	       "\tppc64_cpu --run-mode=<val>    # Set current diagnostics run mode\n\n"
-	       "\tppc64_cpu --frequency         # Determine cpu frequency.\n\n");
+	if (is_subcore_capable())
+		subcores = num_subcores();
+
+	for (i = 0; i < cpus_in_system; i++) {
+
+		if (subcores > 1) {
+			if (i % subcores == 0)
+				printf("Core %3d:\n", i/subcores);
+			printf("  Subcore %3d: ", i);
+		} else {
+			printf("Core %3d: ", i);
+		}
+
+		for (j = 0; j < threads_per_cpu; j++) {
+			thread_num = i*threads_per_cpu + j;
+			online = cpu_online(thread_num) ? '*' : ' ';
+			printf("%4d%c ", thread_num, online);
+		}
+		printf("\n");
+	}
+	return 0;
+}
+
+static void usage(void)
+{
+	printf(
+"Usage: ppc64_cpu [command] [options]\n"
+"ppc64_cpu --smt                     # Get current SMT state\n"
+"ppc64_cpu --smt={on|off}            # Turn SMT on/off\n"
+"ppc64_cpu --smt=X                   # Set SMT state to X\n\n"
+"ppc64_cpu --cores-present           # Get the number of cores present\n"
+"ppc64_cpu --cores-on                # Get the number of cores currently online\n"
+"ppc64_cpu --cores-on=X              # Put exactly X cores online\n\n"
+"ppc64_cpu --dscr                    # Get current DSCR system setting\n"
+"ppc64_cpu --dscr=<val>              # Change DSCR system setting\n"
+"ppc64_cpu --dscr [-p <pid>]         # Get DSCR setting for process <pid>\n"
+"ppc64_cpu --dscr=<val> [-p <pid>]   # Change DSCR setting for process <pid>\n\n"
+"ppc64_cpu --smt-snooze-delay        # Get current smt-snooze-delay setting\n"
+"ppc64_cpu --smt-snooze-delay=<val>  # Change smt-snooze-delay setting\n\n"
+"ppc64_cpu --run-mode                # Get current diagnostics run mode\n"
+"ppc64_cpu --run-mode=<val>          # Set current diagnostics run mode\n\n"
+"ppc64_cpu --frequency [-t <time>]   # Determine cpu frequency for <time>\n"
+"                                    # seconds, default is 1 second.\n\n"
+"ppc64_cpu --subcores-per-core       # Get number of subcores per core\n"
+"ppc64_cpu --subcores-per-core=X     # Set subcores per core to X (1 or 4)\n"
+"ppc64_cpu --info                    # Display system state information)\n");
 }
 
 struct option longopts[] = {
@@ -929,6 +1211,8 @@ struct option longopts[] = {
 	{"frequency",		no_argument,	   NULL, 'f'},
 	{"cores-present",	no_argument,	   NULL, 'C'},
 	{"cores-on",		optional_argument, NULL, 'c'},
+	{"subcores-per-core",	optional_argument, NULL, 'n'},
+	{"info",		no_argument,	   NULL, 'i'},
 	{"version",		no_argument,	   NULL, 'V'},
 	{0,0,0,0}
 };
@@ -936,8 +1220,12 @@ struct option longopts[] = {
 int main(int argc, char *argv[])
 {
 	int rc = 0;
+	char *action;
+	char *action_arg = NULL;
+	char *equal_char;
 	int opt;
-	int option_index;
+	int sleep_time = 1; /* default to one second */
+	pid_t pid = -1;
 
 	if (argc == 1) {
 		usage();
@@ -950,47 +1238,80 @@ int main(int argc, char *argv[])
 		return rc;
 	}
 
+	/* The first arg is the action to be taken with an optional action
+	 * arg in the form --action=XXX. Parse this out so we can call the
+	 * appropriate action.
+	 */
+	action = argv[1];
+
+	/* skipp past the '--' */
+	action += 2;
+
+	equal_char = strchr(action, '=');
+	if (equal_char) {
+		*equal_char = '\0';
+		action_arg = equal_char + 1;
+	}
+
+	/* Now parse out any additional options. */
+	optind = 2;
 	while (1) {
-		opt = getopt_long(argc, argv, "s::d::S::r::fCVc::", longopts,
-				  &option_index);
+		opt = getopt(argc, argv, "p:t:");
 		if (opt == -1)
 			break;
 
 		switch (opt) {
-		    case 's':
-			rc = do_smt(optarg);
-			break;
+		case 'p':
+			/* only valid for do_dscr option */
+			if (strcmp(action, "dscr")) {
+				fprintf(stderr, "The p option is only valid "
+					"with the --dscr option\n");
+				usage();
+				exit(-1);
+			}
 
-		    case 'd':
-			rc = do_dscr(optarg);
+			pid = atoi(optarg);
 			break;
+		case 't':
+			/* only valid for --frequency */
+			if (strcmp(action, "frequency")) {
+				fprintf(stderr, "The t option is only valid "
+					"with the --frequency option\n");
+				usage();
+				exit(-1);
+			}
 
-		    case 'S':
-			rc = do_smt_snooze_delay(optarg);
+			sleep_time = atoi(optarg);
 			break;
-
-		    case 'r':
-			rc = do_run_mode(optarg);
-			break;
-
-		    case 'f':
-			rc = do_cpu_frequency();
-			break;
-
-		    case 'C':
-			rc = do_cores_present(optarg);
-			break;
-		    case 'c':
-			rc = do_cores_online(optarg);
-			break;
-		    case 'V':
-			printf("ppc64_cpu: version %s\n", PPC64_CPU_VERSION);
-			break;
-		    default:
+		default:
+			fprintf(stderr, "%c is not a valid option\n", opt);
 			usage();
-			break;
+			exit(-1);
 		}
 	}
 
+	if (!strcmp(action, "smt"))
+		rc = do_smt(action_arg);
+	else if (!strcmp(action, "dscr"))
+		rc = do_dscr(action_arg, pid);
+	else if (!strcmp(action, "smt-snooze-delay"))
+		rc = do_smt_snooze_delay(action_arg);
+	else if (!strcmp(action, "run-mode"))
+		rc = do_run_mode(action_arg);
+	else if (!strcmp(action, "frequency"))
+		rc = do_cpu_frequency(sleep_time);
+	else if (!strcmp(action, "cores-present"))
+		rc = do_cores_present(action_arg);
+	else if (!strcmp(action, "cores-on"))
+		rc = do_cores_online(action_arg);
+	else if (!strcmp(action, "subcores-per-core"))
+		rc = do_subcores_per_core(action_arg);
+	else if (!strcmp(action, "info"))
+		rc = do_info(action_arg);
+	else if (!strcmp(action, "version"))
+		printf("ppc64_cpu: version %s\n", PPC64_CPU_VERSION);
+	else
+		usage();
+
 	return rc;
 }
